"""
Provide custom exceptions and error handling for the CLI.

Borrowed from Smart Letters CLI: https://github.com/dusktreader/smart-letters
"""

from sys import exc_info
from functools import wraps
from typing import Any, Callable, ParamSpec, TypeVar, final

import buzz
import snick
import typer
from loguru import logger
from rich.console import Console
from rich.panel import Panel


@final
class Abort(buzz.Buzz):
    def __init__(
        self,
        message: str,
        *args: Any,
        subject: str | None = None,
        log_message: str | None = None,
        warn_only: bool = False,
        **kwargs: Any,
    ):
        self.subject = subject
        self.log_message = log_message
        self.warn_only = warn_only
        self.original_error: BaseException | None = None
        (_, self.original_error, __) = exc_info()
        super().__init__(message, *args, **kwargs)


P = ParamSpec("P")
R = TypeVar("R")


def handle_abort(func: Callable[P, R]) -> Callable[P, R]:

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        try:
            return func(*args, **kwargs)
        except Abort as err:
            if not err.warn_only:
                if err.log_message is not None:
                    logger.error(err.log_message)

                if err.original_error is not None:
                    logger.error(f"Original exception: {err.original_error}")

            panel_kwargs: dict[str, Any] = dict()
            if err.subject is not None:
                panel_kwargs["title"] = f"[red]{err.subject}"
            message = snick.dedent(err.message)

            console = Console()
            console.print()
            console.print(Panel(message, **panel_kwargs))
            console.print()
            raise typer.Exit(code=1)

    return wrapper



import sys
from importlib import resources

if sys.version_info >= (3, 12):
    from importlib.resources.abc import Traversable
else:
    from importlib.abc import Traversable


asset_root = resources.files(f"{__package__}.assets")


def asset_path(file_name: str) -> Traversable:
    return asset_root / file_name



import os
from pathlib import Path
from typing import Annotated, Any

from loguru import logger
import typer

from drivel.cli.config import attach_settings
from drivel.cli.constants import EXTRA_THEMES_DIR, ThemeListOutputFormat, ThemeOutputFormat
from drivel.cli.exceptions import Abort, handle_abort
from drivel.cli.extra_themes import attach_extra_themes
from drivel.cli.format import as_spaces, as_lines, as_json, as_yaml, terminal_message
from drivel.cli.utilities import ensure_storage_path
from drivel.themes import Theme


cli = typer.Typer()


@cli.callback(invoke_without_command=True)
def themes(ctx: typer.Context):
    """
    Commands to interact with themes.

    More information can be shown for each sub-command listed below by running it with the
    --help option.
    """
    if ctx.invoked_subcommand is None:
        ctx.get_help()
        terminal_message(
            "No sub-command provided. Please check [bold magenta]usage[/bold magenta]",
            subject="Need an drivel->themes sub-command",
        )
        ctx.exit()


@cli.command("list")
@attach_settings(validate=True)
@attach_extra_themes
@handle_abort
def list_all(
    ctx: typer.Context,
    output_format: Annotated[
        ThemeListOutputFormat, typer.Option("--format", help="The output format to use")
    ] = ThemeListOutputFormat.spaces,
    fancy: Annotated[bool, typer.Option(help="Enable fancy output")] = True,
    to_clipboard: Annotated[bool, typer.Option(help="Copy output to clipboard")] = True,
):
    """
    List all available themes.
    """
    extra_themes_dir: Path = ensure_storage_path(ctx.obj.cache_dir / EXTRA_THEMES_DIR, "extra-themes")
    theme_list = list(Theme.names(extra_themes_dir))
    match output_format:
        case ThemeListOutputFormat.spaces:
            as_spaces(theme_list, fancy=fancy, to_clipboard=to_clipboard)
        case ThemeListOutputFormat.lines:
            as_lines(theme_list, fancy=fancy, to_clipboard=to_clipboard)
        case ThemeListOutputFormat.json:
            as_json(theme_list, fancy=fancy, to_clipboard=to_clipboard)



@cli.command()
@attach_settings(validate=True)
@attach_extra_themes
@handle_abort
def add(
    ctx: typer.Context,
    theme_file: Annotated[Path, typer.Argument(help="The theme file to add")],
):
    """
    Fetch all available themes.
    """
    extra_themes_dir: Path = ctx.obj.extra_themes
    logger.debug(f"Adding theme from file {theme_file}")
    theme = Theme.loads(theme_file.read_text(), name=theme_file.stem)
    theme.dump(extra_themes_dir)
    logger.debug(f"Added theme {theme.name} to {extra_themes_dir}")


@cli.command()
@attach_settings(validate=True)
@attach_extra_themes
@handle_abort
def remove(
    ctx: typer.Context,
    name: Annotated[str, typer.Argument(help="The name of the theme to remove")],
):
    """
    Fetch all available themes.
    """
    extra_themes_dir: Path = ctx.obj.extra_themes
    target: Path = extra_themes_dir / f"{name}.yaml"
    try:
        os.unlink(target)
    except Exception:
        raise Abort(
            f"""
            Could not remove theme {name}.

            It may not have been added or there was a problem removing the file.
            """,
            subject="Remove failed",
            log_message=f"Failed to remove theme file {target}",
        )
    logger.debug(f"Removed {name} from {extra_themes_dir}")


@cli.command()
@attach_settings(validate=True)
@attach_extra_themes
@handle_abort
def show(
    ctx: typer.Context,
    name: Annotated[str, typer.Argument(help="The name of the theme to show")],
    output_format: Annotated[
        ThemeOutputFormat, typer.Option("--format", help="The output format to use")
    ] = ThemeOutputFormat.yaml,
    fancy: Annotated[bool, typer.Option(help="Enable fancy output")] = True,
    to_clipboard: Annotated[bool, typer.Option(help="Copy output to clipboard")] = True,
):
    """
    Show a theme.
    """
    extra_themes_dir: Path = ctx.obj.extra_themes
    try:
        theme = Theme.load(name, extra_themes_dir)
    except Exception:
        raise Abort(
            f"""
            Could not find theme {name}.

            It may not be available or there was a problem accessing the file.
            """,
            subject="Load failed",
            log_message=f"Failed to load theme {name}",
        )

    theme_dict = theme.model_dump()
    match output_format:
        case ThemeOutputFormat.json:
            as_json(theme_dict, fancy=fancy, to_clipboard=to_clipboard)
        case ThemeOutputFormat.yaml:
            as_yaml(theme_dict, fancy=fancy, to_clipboard=to_clipboard)


@cli.command()
@attach_settings(validate=True)
@attach_extra_themes
@handle_abort
def schema(
    ctx: typer.Context,
    fancy: Annotated[bool, typer.Option(help="Enable fancy output")] = True,
    to_clipboard: Annotated[bool, typer.Option(help="Copy output to clipboard")] = True,
):
    """
    Show a theme.
    """
    theme_schema: dict[str, Any] = Theme.model_json_schema()
    as_json(theme_schema, fancy=fancy, to_clipboard=to_clipboard)



from functools import wraps
from typing import Callable, Concatenate, ParamSpec, TypeAlias, TypeVar

import typer

from drivel.cli.constants import CACHE_DIR
from drivel.cli.utilities import attach_to_context, ensure_storage_path

P = ParamSpec("P")
R = TypeVar("R")
WithContext: TypeAlias = Callable[Concatenate[typer.Context, P], R]


def attach_cache_to_context(ctx: typer.Context) -> None:
    attach_to_context(ctx, "cache_dir", ensure_storage_path(CACHE_DIR, "cache"))

def attach_cache(func: WithContext[P, R]) -> WithContext[P, R]:
    @wraps(func)
    def wrapper(ctx: typer.Context, *args: P.args, **kwargs: P.kwargs) -> R:
        attach_cache_to_context(ctx)
        return func(ctx, *args, **kwargs)

    return wrapper



from importlib.abc import Traversable
import re
from pathlib import Path
from random import shuffle
from typing import Annotated

import yaml
from auto_name_enum import AutoNameEnum, auto
from loguru import logger
from pydantic import BaseModel, Field, model_validator

from drivel.constants import DEFAULT_THEME
from drivel.exceptions import ThemeNotFound, ThemeReadError, DuplicateTheme, ThemeWriteError
from drivel.utilities import asset_root


Kinds = dict[str, list[str]]
kind_pattern = r"^[a-z0-9-]+$"


class Display(AutoNameEnum):
    kebab_case = auto()
    snake_case = auto()
    title_case = auto()
    upper_case = auto()


class ThemeMetadata(BaseModel):
    attribution: str | None = None
    explanation: str | None = None


class Theme(BaseModel):
    name: Annotated[str, Field(exclude=True)]
    default: str
    kinds: Kinds
    metadata: ThemeMetadata

    @model_validator(mode="after")
    def validate_theme(self) -> "Theme":
        if self.default not in self.kinds:
            raise ValueError("Default kind not found in kinds")
        for kind, items in self.kinds.items():
            if not re.match(kind_pattern, kind):
                raise ValueError(f"Kind name '{kind}' has invalid characters")
            if not items:
                raise ValueError(f"Kind '{kind}' has no items")
            for item in items:
                if not re.match(kind_pattern, item):
                    raise ValueError(f"Item '{item}' in kind '{kind}' has invalid characters")
        return self

    def give(
        self,
        max_count: int | None = None,
        kind: str | None = None,
        do_shuffle: bool = False,
    ) -> list[str]:
        logger.debug(f"Giving items from {self.name} for kind '{kind}'")
        if kind is None:
            kind = self.default
        if kind == "all":
            items = [i for k in self.kinds.values() for i in k]
        else:
            items = self.kinds[kind]
        if do_shuffle:
            shuffle(items)
        if max_count is None:
            return items
        return items[:max_count]

    @staticmethod
    def _find(name: str, *search_paths: Path | Traversable) -> Path | Traversable:
        for search_path in search_paths:
            for path in search_path.iterdir():
                if path.is_file() and path.name == f"{name}.yaml":
                    return path
        raise ThemeNotFound(f"Theme '{name}' not found in paths: {search_paths}")

    @classmethod
    def loads(cls, text: str, name: str) -> "Theme":
        data = yaml.safe_load(text)
        theme = Theme(name=name, **data)

        return theme

    @classmethod
    def load(cls, name: str = DEFAULT_THEME, *extra_paths: Path) -> "Theme":
        path = cls._find(name, asset_root, *extra_paths)

        with ThemeReadError.handle_errors(f"Couldn't read theme '{name}' from {path}"):
            text = path.read_text()
            theme = cls.loads(text, name)

        return theme

    def dump(self, target: Path) -> None:
        yaml_text = yaml.dump(self.model_dump())
        theme_path = target / f"{self.name}.yaml"
        with ThemeWriteError.handle_errors(f"Couldn't write theme '{self.name}' to {theme_path}"):
            theme_path.write_text(yaml_text)


    @classmethod
    def names(cls, *extra_paths: Path) -> set[str]:
        names: set[str] = set()
        for asset in asset_root.iterdir():
            if not asset.is_file():
                continue
            (name, ext) = asset.name.rsplit('.', 1)
            if ext != "yaml":
                continue
            names.add(name)

        for extra_path in extra_paths:
            for asset in extra_path.iterdir():
                if not asset.is_file():
                    continue
                if asset.suffix != ".yaml":
                    continue
                name = asset.stem
                DuplicateTheme.require_condition(name not in names, f"Duplicate theme name found in {extra_path}: {name}")
                names.add(name)

        return names



from pathlib import Path
from typing import Any

from loguru import logger
import typer

from drivel.cli.exceptions import Abort


def ensure_storage_path(
    storage_path: Path,
    description: str,
) -> Path:
    try:
        storage_path.mkdir(exist_ok=True, parents=True)
        info_file = storage_path / "info.txt"
        info_file.write_text(
            f"""
            This directory is used by Drivel for its {description}.
            """
        )
        return storage_path
    except Exception:
        raise Abort(
            """
            {description.capitalize()} directory {storage_path} doesn't exist, is not writable,
            or could not be created.

            Please check your home directory permissions and try again.
            """,
            subject=f"Non-writable {description} dir",
            log_message=f"Non-writable {description} cache dir",
        )


def attach_to_context(ctx: typer.Context, key: str, value: Any) -> None:
    logger.debug(f"Binding {key} to CLI context")
    if getattr(ctx.obj, key):
        logger.warning(f"Context already has attribute {key}. Skipping.")
        return
    setattr(ctx.obj, key, value)



from functools import wraps
from typing import Callable, Concatenate, ParamSpec, TypeAlias, TypeVar

import typer

from drivel.cli.constants import SHARE_DIR
from drivel.cli.utilities import attach_to_context, ensure_storage_path

P = ParamSpec("P")
R = TypeVar("R")
WithContext: TypeAlias = Callable[Concatenate[typer.Context, P], R]


def attach_share_to_context(ctx: typer.Context) -> None:
    attach_to_context(ctx, "share_dir", ensure_storage_path(SHARE_DIR, "share"))


def attach_share(func: WithContext[P, R]) -> WithContext[P, R]:
    @wraps(func)
    def wrapper(ctx: typer.Context, *args: P.args, **kwargs: P.kwargs) -> R:
        attach_share_to_context(ctx)
        return func(ctx, *args, **kwargs)

    return wrapper



from pathlib import Path
from typing import Annotated, Any

import typer

from drivel.exceptions import DrivelError
from drivel.cli.config import Settings, attach_settings
from drivel.cli.constants import ThemeListOutputFormat
from drivel.cli.exceptions import Abort, handle_abort
from drivel.cli.extra_themes import attach_extra_themes
from drivel.cli.format import as_spaces, as_lines, as_json
from drivel.themes import Theme


cli = typer.Typer()


@cli.callback(invoke_without_command=True)
@attach_settings(validate=True)
@attach_extra_themes
@handle_abort
def give(
    ctx: typer.Context,
    max_count: Annotated[int | None, typer.Argument(help="The maximum number of metasyntactic names to give")] = None,
    do_shuffle: Annotated[bool, typer.Option("--shuffle", help="Mix the names")] = False,
    theme_name: Annotated[
        str | None, typer.Option("--theme", help="The theme to use (If not provided, will use current default")
    ] = None,
    kind: Annotated[str | None, typer.Option(help="The kind of names to give. Use 'all' to pull from all kinds")] = None,
    output_format: Annotated[
        ThemeListOutputFormat, typer.Option("--format", help="The output format to use")
    ] = ThemeListOutputFormat.spaces,
    fancy: Annotated[bool, typer.Option(help="Enable fancy output")] = True,
    to_clipboard: Annotated[bool, typer.Option(help="Copy output to clipboard")] = True,
):
    """
    Give N fun metasyntactic variable names.
    """
    settings: Settings = ctx.obj.settings
    extra_themes_dir: Path = ctx.obj.extra_themes
    if theme_name is None:
        theme_name = settings.default_theme

    try:
        theme = Theme.load(theme_name, extra_themes_dir)
        items = theme.give(max_count=max_count, kind=kind, do_shuffle=do_shuffle)
    except DrivelError:
        raise Abort(
            "There was an error fetching names!",
            subject="Error giving names",
        )
    kwargs: dict[str, Any] = dict(
        fancy=fancy,
        to_clipboard=to_clipboard,
    )
    match output_format:
        case ThemeListOutputFormat.spaces | ThemeListOutputFormat.lines:
            if fancy:
                kwargs["subject"] = f"Fun names from {theme_name}"
            match output_format:
                case ThemeListOutputFormat.spaces:
                    as_spaces(items, **kwargs)
                case ThemeListOutputFormat.lines:
                    as_lines(items, **kwargs)
        case ThemeListOutputFormat.json:
            as_json(items, **kwargs)



from typing import Annotated
import typer

from drivel.cli.config import cli as config_cli
from drivel.cli.format import terminal_message
from drivel.cli.give import cli as give_cli
from drivel.cli.logging import init_logs
from drivel.cli.schemas import CliContext
from drivel.cli.themes import cli as themes_cli
from drivel.cli.version import show_version


cli = typer.Typer(rich_markup_mode="rich")


@cli.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    verbose: Annotated[bool, typer.Option(help="Enable verbose logging to the terminal")] = False,
    version: Annotated[bool, typer.Option(help="Print the version of this app and exit")] = False,
):
    """
    Welcome to drivel!

    More information can be shown for each command listed below by running it with the
    --help option.
    """

    if version:
        show_version()
        ctx.exit()

    if ctx.invoked_subcommand is None:
        ctx.get_help()
        terminal_message(
            "No command provided. Please check [bold magenta]usage[/bold magenta]",
            subject="Need an drivel command",
        )
        ctx.exit()

    init_logs(verbose=verbose)
    ctx.obj = CliContext()


cli.add_typer(config_cli, name="config")
cli.add_typer(give_cli, name="give")
cli.add_typer(themes_cli, name="themes")


if __name__ == "__main__":
    cli()



from dataclasses import dataclass
from pathlib import Path

from drivel.cli.config import Settings


@dataclass
class CliContext:
    settings: Settings | None = None
    cache_dir: Path | None = None
    share_dir: Path | None = None
    extra_themes: Path | None = None



from pathlib import Path

from auto_name_enum import AutoNameEnum, auto


class ThemeListOutputFormat(AutoNameEnum):
    json = auto()
    spaces = auto()
    lines = auto()


class ThemeOutputFormat(AutoNameEnum):
    json = auto()
    yaml = auto()


SHARE_DIR: Path = Path.home() / ".local/share/drivel"
CONFIG_DIR: Path = Path.home() / ".config/drivel"
CACHE_DIR: Path = Path.home() / ".cache/drivel"

SETTINGS_FILE_NAME: str = "settings.json"
EXTRA_THEMES_DIR: str = "extra-themes"



"""
Provide custom exceptions.
"""

import buzz
from rich import traceback


# Enables prettified traceback printing via rich
traceback.install()


class DrivelError(buzz.Buzz):
    pass


class ThemeNotFound(DrivelError):
    pass


class ThemeReadError(DrivelError):
    pass


class ThemeWriteError(DrivelError):
    pass


class DuplicateTheme(DrivelError):
    pass



import json
from typing import Any
from collections.abc import Iterable

import pyperclip
import snick
from loguru import logger
from rich.console import Console
from rich.panel import Panel
from rich.markdown import Markdown
import yaml


def _to_clipboard(text: str) -> bool:
    try:
        pyperclip.copy(text)
        return True
    except Exception as exc:
        logger.debug(f"Could not copy letter to clipboard: {exc}")
        return False


def terminal_message(
    message: str,
    subject: str | None = None,
    color: str = "green",
    footer: str | None = None,
    indent: bool = True,
    markdown: bool = False,
    to_clipboard: bool = False,
):
    if to_clipboard:
        result = _to_clipboard(message)
        if result and not footer:
            footer = "Copied to clipboard!"
    panel_kwargs: dict[str, Any] = dict(padding=1)
    if subject is not None:
        panel_kwargs["title"] = f"[{color}]{subject}"
    if footer is not None:
        panel_kwargs["subtitle"] = f"[dim italic]{footer}[/dim italic]"
    text = snick.dedent(message)
    if indent:
        text = snick.indent(text, prefix="  ")
    console = Console()
    console.print()
    if markdown:
        console.print(Panel(Markdown(text), **panel_kwargs))
    else:
        console.print(Panel(text, **panel_kwargs))
    console.print()


def simple_message(
    message: str,
    indent: bool = False,
    markdown: bool = False,
    to_clipboard: bool = False,
):
    text = snick.dedent(message)
    if indent:
        text = snick.indent(text, prefix="  ")
    if to_clipboard:
        result = _to_clipboard(text)
        if result:
            logger.debug("output copied to clipboard")
    console = Console()
    console.print()
    if markdown:
        console.print(Markdown)
    else:
        console.print(text)
    console.print()


def as_spaces(stuff: Iterable[Any], fancy: bool = False, **kwargs: Any):
    output = " ".join(stuff)
    if fancy:
        return terminal_message(output, **kwargs)
    else:
        kwargs.pop("subject", None)
    return simple_message(output, **kwargs)


def as_lines(stuff: Iterable[Any], fancy: bool = False, **kwargs: Any):
    output = "\n".join(stuff)
    if fancy:
        return terminal_message(output, **kwargs)
    else:
        kwargs.pop("fancy", None)
    return simple_message(output, **kwargs)


def as_json(stuff: Any, fancy: bool = False, to_clipboard: bool = False, **kwargs: Any):
    indent: int | None = 2 if fancy else None
    if to_clipboard:
        result = _to_clipboard(json.dumps(stuff, indent=indent))
        if result:
            logger.debug("output copied to clipboard")
    console = Console()
    console.print()
    console.print_json(data=stuff, indent=indent, **kwargs)
    console.print()


def as_yaml(stuff: Any, fancy: bool = False, to_clipboard: bool = False, **kwargs: Any):
    if fancy:
        logger.warning("as_yaml doesn't support fancy output")
    text = yaml.dump(stuff, sort_keys=False)
    if to_clipboard:
        result = _to_clipboard(text)
        if result:
            logger.debug("output copied to clipboard")
    console = Console()
    console.print()
    console.print(text, **kwargs)
    console.print()



from contextlib import contextmanager
import json
from functools import wraps
from pathlib import Path
from typing import Annotated, Any, Callable, Concatenate, ParamSpec, TypeAlias, TypeVar, cast

import snick
import typer
from inflection import dasherize
from loguru import logger
from pydantic import BaseModel, ValidationError, Field

from drivel.cli.share import attach_share, attach_share_to_context
from drivel.cli.utilities import attach_to_context
from drivel.constants import DEFAULT_THEME
from drivel.cli.constants import SETTINGS_FILE_NAME
from drivel.cli.exceptions import Abort, handle_abort
from drivel.cli.format import terminal_message


P = ParamSpec("P")
R = TypeVar("R", covariant=True)
WithContext: TypeAlias = Callable[Concatenate[typer.Context, P], R]


def file_exists(value: Path | None) -> Path | None:
    if value is None:
        return value

    value = value.expanduser()
    if not value.exists():
        raise ValueError(f"File not found at {value}")
    return value


class Settings(BaseModel):
    default_theme: str = DEFAULT_THEME

    invalid_warning: Annotated[
        str | None,
        Field(
            exclude=True,
            description="""
            An optional warning that can be included when the model is invalid.

            Used when we use the `attach_settings` decorator with `validate=False`.
        """,
        ),
    ] = None


@contextmanager
def handle_config_error():
    try:
        yield
    except ValidationError as err:
        raise Abort(
            snick.conjoin(
                "A configuration error was detected.",
                "",
                "Details:",
                "",
                f"[red]{err}[/red]",
            ),
            subject="Configuration Error",
            log_message="Configuration error",
        )


def init_settings(validate: bool = True, **settings_values: Any) -> Settings:
    with handle_config_error():
        logger.debug("Validating settings")
        try:
            return Settings(**settings_values)
        except ValidationError as err:
            if validate:
                raise
            settings = Settings.model_construct(**settings_values)
            settings.invalid_warning = str(err)
            return settings


def update_settings(settings: Settings, **settings_values: Any) -> Settings:
    with handle_config_error():
        logger.debug("Validating settings")
        settings_dict = settings.model_dump(exclude_unset=True)
        settings_dict.update(**settings_values)
        return Settings(**settings_dict)


def unset_settings(settings: Settings, *unset_keys: str) -> Settings:
    with handle_config_error():
        logger.debug("Unsetting settings")
        return Settings(**{k: v for (k, v) in settings.model_dump(exclude_unset=True).items() if k not in unset_keys})


def attach_settings(validate: bool = True) -> Callable[[WithContext[P, R]], WithContext[P, R]]:
    """
    Attach the settings to the CLI context.

    If the share directory hasn't been attached yet, attach it.

    Optionally, skip validation of the settings. This is useful in case the config
    file being loaded is not valid, but we still want to use the settings. Then, we
    can update the settings with correct values.

    I would love to figure out how to make this work with optional parameters as described in:
    https://stackoverflow.com/a/24617244/642511

    However, I couldn't figure out the typing for it.
    """

    def _decorate(func: WithContext[P, R]) -> WithContext[P, R]:

        @wraps(func)
        def wrapper(ctx: typer.Context, *args: P.args, **kwargs: P.kwargs) -> R:
            attach_share_to_context(ctx)
            settings_path = ctx.obj.share_dir / "settings.json"

            settings_values: dict[str, Any] = {}
            try:
                logger.debug(f"Loading settings from {settings_path}")
                settings_values.update(**json.loads(settings_path.read_text()))
            except FileNotFoundError:
                # If I ever port this into a cli toolkit, need a way to detect if all settings have a default and only
                # trigger this if they do not
                #
                # raise Abort(
                #     f"""
                #     No settings file found at {settings_path}!

                #     Run the bind sub-command first to establish your settings.
                #     """,
                #     subject="Settings file missing!",
                #     log_message="Settings file missing!",
                # )
                pass
            settings: Settings = init_settings(validate=validate, **settings_values)
            attach_to_context(ctx, "settings", settings)
            return func(ctx, *args, **kwargs)

        return wrapper

    return _decorate


def dump_settings(settings: Settings, settings_path: Path) -> None:
    logger.debug(f"Saving settings to {settings_path}")
    settings_values = settings.model_dump_json(indent=2)
    settings_path.write_text(settings_values)


def clear_settings(settings_path: Path) -> None:
    logger.debug(f"Removing saved settings at {settings_path}")
    settings_path.unlink(missing_ok=True)


def show_settings(settings: Settings):
    parts: list[tuple[str, str]] = []
    for field_name, field_value in settings:
        if field_name == "invalid_warning":
            continue
        parts.append((dasherize(field_name), field_value))
    max_field_len = max(len(field_name) for field_name, _ in parts)
    message = "\n".join(f"[bold]{k:<{max_field_len}}[/bold] -> {v}" for k, v in parts)
    if settings.invalid_warning:
        message += f"\n\n[red]Configuration is invalid: {settings.invalid_warning}[/red]"
    terminal_message(message, subject="Current Configuration")


cli = typer.Typer(help="Configure the app, change settings, or view how it's currently configured")


@cli.command()
@attach_share
@handle_abort
def bind(
    ctx: typer.Context,
    default_theme: Annotated[str | None, typer.Option(help="The default theme to use in the CLI")] = None,
) -> None:
    """
    Bind the configuration to the app.
    """
    filtered_locals: dict[str, Any] = {k: v for (k, v) in locals().items() if k != "ctx" and v is not None}
    logger.debug(f"Initializing settings with {filtered_locals}")
    settings = init_settings(**filtered_locals)

    settings_path: Path = cast(Path, ctx.obj.share_dir / SETTINGS_FILE_NAME)
    dump_settings(settings, settings_path)
    show_settings(settings)


@cli.command()
@attach_settings(validate=False)
@handle_abort
def update(
    ctx: typer.Context,
    default_theme: Annotated[str | None, typer.Option(help="The default theme to use in the CLI")] = None,
) -> None:
    """
    Update one or more configuration settings that are bound to the app.
    """
    filtered_locals: dict[str, Any] = {k: v for (k, v) in locals().items() if k != "ctx" and v is not None}
    logger.debug(f"Updating settings with {filtered_locals}")
    settings = update_settings(ctx.obj.settings or Settings(), **filtered_locals)

    settings_path: Path = cast(Path, ctx.obj.share_dir / SETTINGS_FILE_NAME)
    dump_settings(settings, settings_path)
    show_settings(settings)


@cli.command()
@handle_abort
@attach_settings(validate=False)
def unset(
    ctx: typer.Context,
    default_theme: Annotated[bool, typer.Option(help="The default theme to use in the CLI")] = False,
):
    """
    Remove a configuration setting that was previously bound to the app.
    """
    keys: list[str] = [k for (k, v) in locals().items() if k != "ctx" and v]
    logger.debug(f"Unsetting settings: {keys}")
    settings = unset_settings(ctx.obj.settings, *keys)

    settings_path: Path = cast(Path, ctx.obj.share_dir / SETTINGS_FILE_NAME)
    dump_settings(settings, settings_path)
    show_settings(settings)


@cli.command()
@handle_abort
@attach_settings(validate=False)
def show(ctx: typer.Context):
    """
    Show the config that is currently bound to the app.
    """
    settings: Settings = ctx.obj.settings or Settings()
    show_settings(settings)


@cli.command()
@handle_abort
@attach_settings(validate=False)
def path(ctx: typer.Context):
    """
    Show the path to the config file that is currently bound to the app.
    """
    settings_path: Path = cast(Path, ctx.obj.share_dir / SETTINGS_FILE_NAME)
    terminal_message(str(settings_path), subject="Current Configuration Path")


@cli.command()
@handle_abort
@attach_share
def clear(ctx: typer.Context):
    """
    Clear the config from the app.
    """
    doit = typer.confirm("Are you sure you want to clear the settings?")
    if doit:
        logger.debug("Clearing settings")
        settings_path: Path = cast(Path, ctx.obj.share_dir / SETTINGS_FILE_NAME)
        clear_settings(settings_path)
        terminal_message(
            "All settings have been cleared and returned to built-in defaults",
            subject="Settings Cleared"
        )
    else:
        logger.debug("Clearing settings aborted")
